
/*
 *	arch/arc/drivers/arc-ide.c
 *
 *	ARC IDE Interface 'chipset' functions
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *		 
 *	Copyright 2005,2006 ARC International
 *
 */

#include <linux/ide.h>
#include <asm/ide.h>
#include <asm/delay.h>

//#define DEBUG

#ifdef DEBUG 
#define DBG(fmt, args...) printk(KERN_INFO fmt , ## args)
#else
#define DBG(fmt, args...)
#endif

/* Need to know the system clock frequency for timing set-up */
extern unsigned long arc700_clkfreq;

/* Timing parameters. These are the figures taken from Joe's
	 document "Aurora IDE Interface Implementation Specification".
	 This is arranged as an array of 7 parameters for each of the
	 5 modes that we support. The individual timing parameters are
	 (in turn):
	 - T0	 Minimum PIO cycle time
	 - T1	 Minimum addr setup before nDIOR/W
	 - T2	 Minimum nDIOR/W pulse width
	 - T2L Minimum nDIOR/W recovery time (modes 3&4 only)
	 - T3	 Minimum write data setup time
	 - T4	 Minimum write data hold time
	 - T5	 Minimum read data setup time
	 - T6	 Minimum read data hold data
	 All figures are in nanoseconds.
*/
static unsigned long timings[5][8] = {
	{ 600, 70, 165, 0,	60, 30, 50, 5 },
	{ 383, 50, 125, 0,	45, 20, 35, 5 },
	{ 240, 30, 100, 0,	30, 15, 20, 5 },
	{ 180, 30, 80,	70, 30, 10, 20, 5 },
	{ 120, 25, 70,	25, 20, 10, 20, 5 }
	//{ 120, 25, 100,	25, 20, 10, 20, 5 }
}; 

/* We're actually only interested in these. Why did I just type the rest in ? */
#define TIMING_T0 0
#define TIMING_T1 1
#define TIMING_T2 2
#define TIMING_T2L 3

/* Need to convert times to clock cycles and populate register */
static double cycles_per_nanosec;
static __inline__ unsigned long ns_to_cycles(unsigned long t) 
{ 
	unsigned long tmp; 
	tmp = (unsigned long) (cycles_per_nanosec * ((double) t)); 
	//	return tmp ? tmp : 0xff; 
	return tmp;
}
#define TIMING_REG_FORMAT(t1,t2,t2l) ((((t2l) & 0xff) << 16) | (((t2) & 0xff) << 8) | ((t1) & 0xff))

/* New ARC ide method method */
#define USE_NEW_TIMING_METHOD 1
#define NS2CYCLES(X) ((double)(cycles_per_nanosec*(X)))


/* Here's the controller */
static volatile ARC_IDE_if *controller = (volatile ARC_IDE_if*) IDE_CONTROLLER_BASE;

static int initialising;

/* --- Part 1: IDE interface standard functions --- */

/* routine is for HBA specific IDENTITY operations */
int arc_ide_identify(ide_drive_t *drive)
{
	DBG("arc_ide_identify(ide_drive_t *)\n");
	return 0;
}

/*******************************************************************************
*  FUN     : _arc_ide_get_timing_                                              *
*  INPUT   : unsigned float                                                    *
*  OUTPUT  : unsigned long                                                     *
*******************************************************************************/
unsigned long arc_ide_get_timing(double _clock) {
     unsigned long _safe_clock = 1;
     if( _clock >= 255 ) {
          return 255;     
     }
     while( _safe_clock < _clock ) {
          _safe_clock++;        
     }
     return _safe_clock;
}

/* routine to tune PIO mode for drives */
void arc_ide_tuneproc(ide_drive_t *drive, unsigned char pio_mode)
{
	unsigned long t0,t1,t2,t2l;

	DBG("arc_ide_tuneproc(ide_drive_t *, u8) - drive %s, mode %d\n",drive->name,pio_mode);

/* 	if (! initialising) */
/* 	  pio_mode = ide_get_best_pio_mode(drive, pio_mode, 4, &d); */

	if (pio_mode > 4) pio_mode = 4;

	cycles_per_nanosec = ((double)arc700_clkfreq)/1000000000.0;
#if USE_NEW_TIMING_METHOD // jinn - for better timming parameter
     t0 = arc_ide_get_timing(NS2CYCLES(timings[pio_mode][TIMING_T0]));
	
	t1 = arc_ide_get_timing(NS2CYCLES(timings[pio_mode][TIMING_T1]));
	
	t2 = arc_ide_get_timing(NS2CYCLES(timings[pio_mode][TIMING_T2]));
	
	t2l = arc_ide_get_timing(NS2CYCLES(timings[pio_mode][TIMING_T2L]));
#else
	t0 = ns_to_cycles(timings[pio_mode][TIMING_T0]);
	t1 = ns_to_cycles(timings[pio_mode][TIMING_T1]);
	t2 = ns_to_cycles(timings[pio_mode][TIMING_T2]);
	t2l = ns_to_cycles(timings[pio_mode][TIMING_T2L]);
#endif
	DBG("t0 is %d, t1 is %d, t2 is %d, t2l is %d\n",t0,t1,t2,t2l);

	while ((t1+t2+t2l) < t0) { 
		DBG("Incrementing to match basic timing requirements\n");
		//		t1++; t2++; t2l++;
		t2++;
		if ((t1+t2+t2l) < t0) t2l++; 
		if ((t1+t2+t2l) < t0) t1++; 
	}

	controller->timing = TIMING_REG_FORMAT(t1,t2,t2l);
	DBG("arc_ide_tuneproc(u8) - timing reg is now 0x%x\n",controller->timing);
#if 1 // jinn - 20060828 - show current pio timming
     printk(KERN_INFO "arc_ide_tuneproc(u8) - timing reg is now 0x%x\n",controller->timing);
     printk(KERN_INFO "arc_ide_tuneproc(u8) - pio mode %d\n",pio_mode);
#endif

	return;
}

/* Lots of dummy functions for debugging ... */
/* routine to retune DMA modes for drives */
int arc_ide_speedproc(ide_drive_t *drive, u8 chararg)
{
	DBG("arc_ide_speedproc(ide_drive_t *, u8)\n");
	/* We don't do DMA */
	return 0;
}

/* tweaks hardware to select drive - Not used */
void arc_ide_selectproc(ide_drive_t *drive)
{
//	int i;
	DBG("**** arc_ide_selectproc(ide_drive_t *), drive %s\n", drive->name);

	if (drive->name[2] == 'a')
	{
		drive->present=1;
	}
}

/* chipset polling based on hba specifics */
int arc_ide_reset_poll(ide_drive_t *drive)
{
	DBG("**** arc_ide_reset_poll(ide_drive_t *)\n");
	return 0;
}

/* chipset specific changes to default for device-hba resets */
void arc_ide_pre_reset(ide_drive_t *drive)
{
	DBG("**** arc_ide_pre_reset(ide_drive_t *)\n");
}

/* routine to reset controller after a disk reset */
void arc_ide_resetproc(ide_drive_t *drive)
{
//	int i;

	DBG("**** arc_ide_resetproc(ide_drive_t *)\n");
	controller->statctrl = IDE_STATCTRL_RS | IDE_STATCTRL_IC;
	udelay(5);
	controller->statctrl = IDE_STATCTRL_IE;
}

/* special interrupt handling for shared pci interrupts */
void arc_ide_intrproc(ide_drive_t *drive)
{
	DBG("**** arc_ide_intrproc(ide_drive_t *)\n");
	controller->statctrl = IDE_STATCTRL_IC;
}

/* special host masking for drive selection */
void arc_ide_maskproc(ide_drive_t *drive, int intarg)
{
	DBG("**** arc_ide_maskproc(ide_drive_t *, int)\n");
}

/* check host's drive quirk list */
int arc_ide_quirkproc(ide_drive_t *drive)
{
	DBG("**** arc_ide_quirkproc(ide_drive_t *)\n");
	return 0;
}

/* driver soft-power interface */
int arc_ide_busproc(ide_drive_t *drive, int intarg)
{
	DBG("**** arc_ide_busproc(ide_drive_t *, int)\n");
	return 0;
}

int arc_ide_dma_on(ide_drive_t *drive)
{
	drive->using_dma = 1;
	return 0;
}

int arc_ide_dma_off(ide_drive_t *drive)
{
	drive->using_dma = 0;
	return 0;
}

int arc_ide_dma_check(ide_drive_t *drive)
{
	return 0;
}

int arc_ide_dma_read(ide_drive_t *drive)
{
	ide_hwif_t *hwif	= HWIF(drive);
	struct request *rq	= HWGROUP(drive)->rq;
	DBG("**** arc_ide_read(ide_drive_t *)\n");
	return 0;
}

int arc_ide_dma_write(ide_drive_t *drive)
{
	DBG("**** arc_ide_write(ide_drive_t *)\n");
	return 0;
}
#if 0
int arc_ide_does_not_do_dma(ide_drive_t *drive)
{
	printk(KERN_INFO "\n\n********* Attempt to use DMA function on IDE drive %s *********\n\n",drive->name);
	return -1; /* A bit draconian ? */
}
#endif
int arc_enable_ide_intr()
{
	unsigned long reg;

	reg = read_new_aux_reg(AUX_IENABLE);
	reg |= 1<<ARC_IDE_INTERRUPT;
	write_new_aux_reg(AUX_IENABLE, reg);

	controller->statctrl |= IDE_STATCTRL_IE;

	return 0;
}


/* Not used any more */
int arc_ide_reset_intr()
{
//	int i;

	controller->statctrl |= IDE_STATCTRL_IC;

	while (controller->statctrl & IDE_STATCTRL_IS)
	{
		//		printk("Attempting to clear IDE interrupt\n");
		controller->statctrl |= IDE_STATCTRL_IC;
		udelay(1);
	}

	controller->statctrl &= ~IDE_STATCTRL_IC;

	return 0;
}

int dump_arc_ide()
{
	printk("ARC IDE status:\n");
	if (controller->statctrl & IDE_STATCTRL_IS) printk(" Interrupt active\n"); else printk(" Interrupt clear\n");
	if (controller->statctrl & IDE_STATCTRL_IE) printk(" Interrupt enabled\n"); else printk(" Interrupt disabled\n");

	return 0;
}


/* --- Part 2: Module code --- */

#ifndef MODULE

void __init arc_ide_init(void)
{
//	int i;
	unsigned long reg;

	initialising = 1;

	printk(KERN_INFO "ARC IDE interface driver, controller version %d\n",((controller->ID>>16) & 0xff));

#if 1 // jinn - 20060828 - set gpio 10 and gpio 11 status
     printk(KERN_INFO "Initial gpio state, to disable buffer switch\n");
     (*(volatile unsigned int*)(0xC1FC0000)) &= 0xFFFFFCFF;// GPIO_DIR
     (*(volatile unsigned int*)(0xC1FC0000)) |= 0x0000FC00;// GPIO_DIR
     (*(volatile unsigned int*)(0xC1FC0004)) |= 0x0000FC00;// GPIO_DATA
     printk(KERN_INFO "GPIO_REG[0] = %08x\n", (*(volatile unsigned int*)(0xC1FC0000)));
     printk(KERN_INFO "GPIO_REG[1] = %08x\n", (*(volatile unsigned int*)(0xC1FC0004)));
#endif

#ifndef MAXTOR // Reset behaviour of Seagates and Maxtors is different.
	controller->statctrl = 0;
	udelay(50);
	controller->statctrl = IDE_STATCTRL_RS;
	udelay(50);
#endif

	reg = read_new_aux_reg(AUX_ITRIGGER);
	reg |= 1<<ARC_IDE_INTERRUPT;
	write_new_aux_reg(AUX_ITRIGGER, reg);

	reg = read_new_aux_reg(AUX_IENABLE);
	reg &= ~(1<<ARC_IDE_INTERRUPT);
	write_new_aux_reg(AUX_IENABLE, reg);

/*	 controller->statctrl = 0; */
/*	 //	 for (i = 0; i < 10000; i++); */
/*	 udelay(5); */
#ifndef MAXTOR
	controller->statctrl = IDE_STATCTRL_IC|IDE_STATCTRL_RS;
#else
	controller->statctrl = IDE_STATCTRL_IC;
#endif

	udelay(5);

#ifdef MAXTOR
	for (i = 0; i < 6210000; i++) // 31 seconds timeout
	{
	  reg = *((unsigned long*)(IDE_CONTROLLER_BASE + DRIVE_STATCTRL_OFFSET));
	  if (reg & 0x80)
	    udelay(5000);
	  else
	    break;
	}

	if (i >= 1000) printk(KERN_INFO "ARC IDE interface driver, gave up waiting for drive busy to clear\n");
#endif

	// Not sure how correct this is, but it works for me.
	if (arc700_clkfreq >= 70000000)
	  arc_ide_tuneproc(0,4);
	else
	  arc_ide_tuneproc(0,2);
	ide_hwifs[0].chipset = ide_unknown;
	ide_hwifs[0].tuneproc = arc_ide_tuneproc;
	ide_hwifs[0].mmio = 1;

//	ide_hwifs[0].ide_dma_on = arc_ide_dma_on;
//	ide_hwifs[0].ide_dma_off = arc_ide_dma_off;
//	ide_hwifs[0].ide_dma_check = arc_ide_dma_check;
//	ide_hwifs[0].ide_dma_read = arc_ide_dma_read;
//	ide_hwifs[0].ide_dma_write = arc_ide_dma_write;
#ifdef DEBUG
	/* Fill the rest in for tracing */
	ide_hwifs[0].identify = 0;
	ide_hwifs[0].speedproc = 0;
	ide_hwifs[0].selectproc = 0;
	ide_hwifs[0].reset_poll = 0;
	ide_hwifs[0].pre_reset = 0;
	ide_hwifs[0].resetproc = 0;
	ide_hwifs[0].intrproc = 0;
	ide_hwifs[0].maskproc = 0;
	ide_hwifs[0].quirkproc = 0;
	ide_hwifs[0].busproc = 0;

	ide_hwifs[0].ide_dma_read = 0;
	ide_hwifs[0].ide_dma_write = 0;
	ide_hwifs[0].ide_dma_begin = 0;
	ide_hwifs[0].ide_dma_end = 0;
	ide_hwifs[0].ide_dma_check = 0;
	ide_hwifs[0].ide_dma_on = 0;
	ide_hwifs[0].ide_dma_off = 0;
	ide_hwifs[0].ide_dma_off_quietly = 0;
	ide_hwifs[0].ide_dma_test_irq = 0;
	ide_hwifs[0].ide_dma_host_on = 0;
	ide_hwifs[0].ide_dma_host_off = 0;
	ide_hwifs[0].ide_dma_bad_drive = 0;
	ide_hwifs[0].ide_dma_good_drive = 0;
	ide_hwifs[0].ide_dma_count = 0;
	ide_hwifs[0].ide_dma_verbose = 0;
	ide_hwifs[0].ide_dma_retune = 0;
	ide_hwifs[0].ide_dma_lostirq = 0;
	ide_hwifs[0].ide_dma_timeout = 0;
#endif

	initialising = 0;
}

#else

MODULE_AUTHOR("ARC International");
MODULE_DESCRIPTION("ARC IDE interface");
MODULE_LICENSE("GPL");

void __init arc_ide_mod_init(void)
{
	arc_ide_init();
}
module_init(arc_ide_mod_init);

void __init arc_ide_exit(void)
{
	controller->statctrl = 0;
	ide_hwifs[0].chipset = ide_unknown;
	ide_hwifs[0].tuneproc = NULL;
}
module_exit(arc_ide_mod_exit);

#endif

