/*
 * BRIEF MODULE DESCRIPTION
 *	ARC Video Syncronisation Utility
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of	the GNU General	Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * THIS	SOFTWARE IS PROVIDED``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.	 IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the	 GNU General Public License along
 * with this program; if not, write	 to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 2004-2006 ARC International
 */

#include "arc_vsync.h"
#include <linux/kfifo.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include "arc_pgu.h"

#define DEFAULT_FRAME_RATE 30

/*Flowbee*/
#define DISP_START_BUF		0			
static int ref_cnt[CONFIG_ARCPGU_YUVBUFS]; 		/* reference count for buffers */
static int dec_ref_cnt[CONFIG_ARCPGU_YUVBUFS]; 	/* Flowbee: decode buffer index */
extern unsigned char scalar_num_buffers;	// mplayer allocate the scalar buf nums.
extern unsigned char dec_num_buffers;
extern unsigned char disp_num_buffers;
extern unsigned char sw_osd_flag;
extern unsigned char is_video_play;
extern int mplayer_status;

static struct kfifo *display_q; /* queue of buffers to be displayed*/
static struct kfifo *done_q; /* queue of buffers to be freed */
static double frame_rate = DEFAULT_FRAME_RATE;
static int frame_delay = HZ/DEFAULT_FRAME_RATE; 
static double exact_frame_delay = (double)HZ/DEFAULT_FRAME_RATE;
static int run_len = 0;
static unsigned long run_start;
static int is_paused = 0;
static int is_framerate_changed = 0;
static unsigned long last_frame_timestamp = 0; 

/* callback from driver to display buf on HW */
static void (*do_switch)(int buf); 

/* queue of processes waiting for buffer allocation */
static DECLARE_WAIT_QUEUE_HEAD(blocked_q); 

static struct timer_list flip_timer;

static struct timer_list pause_timer;

#ifdef CONFIG_PROC_FS

/* Stuff for /proc/fb_sync */
#define PROC_NAME "fb_sync"
#define TRACE_LEN 64 /* must be power of 2 */
static struct log_t {
    	int buf_idx;
	unsigned long jif;
	int qlen;
} trace_log[TRACE_LEN];

#define TRACE_NEXT(i) (((i)+1) & (TRACE_LEN - 1))

static int trace_log_freep = 0;

static int sync_read_cb(char *buf, char **start, off_t offset,
				  int length, int *eof, void *data)
{
	int i, len=0, freep = trace_log_freep;
	unsigned long last = 0;

	len += sprintf(buf+len, "display_q len = %d\n", __kfifo_len(display_q));
	len += sprintf(buf+len, "Ref_cnt: ");
	for (i=0; i<disp_num_buffers; i++)
	    	len += sprintf(buf+len, " %2d", ref_cnt[i]);
	len += sprintf(buf+len, "\n");

	len += sprintf(buf+len, "%3s %8s %4s %s\n", "idx", "time", "diff", 
		"qlen");
	for (i = TRACE_NEXT(freep); i != freep; i = TRACE_NEXT(i))
	{
	    	unsigned long j = trace_log[i].jif;
		int x = trace_log[i].buf_idx;
		int q = trace_log[i].qlen;

		if (!j)
		    	continue;
		if (last)
			len += sprintf(buf+len, "%3d %8lu %4lu %d\n", 
				x, j, j - last, q);
		last = j;
	}
	if (is_paused)
		len += sprintf(buf+len, "paused\n");

	len += sprintf(buf+len, "frame_rate=%d/1000\n", (int)(1000*frame_rate));

	*eof = 1;
	return len;
}

#endif

static void vsync_reset(void)
{
	int i;

	del_timer(&flip_timer);
	del_timer(&pause_timer);
	__kfifo_reset(display_q);
	__kfifo_reset(done_q);
	is_paused = 0;
	//-- Flowbee 2008-05-14 mplayer not start now, it will start after buffer ready.
	is_video_play = 0;
	//-- Flowbee 2008-05-20 
	mplayer_status = MPLAYER_INIT;
	
	for (i=0; i<CONFIG_ARCPGU_YUVBUFS; i++)
		ref_cnt[i] = 0;

	for (i=0; i<CONFIG_ARCPGU_YUVBUFS; i++)
		dec_ref_cnt[i] = 0;

#ifdef CONFIG_PROC_FS
	memset(trace_log, 0, sizeof(trace_log));
#endif
}

static void log_trace(int buf)
{
#ifdef CONFIG_PROC_FS
 	trace_log[trace_log_freep].buf_idx = buf;
	trace_log[trace_log_freep].jif = jiffies;
	trace_log[trace_log_freep].qlen = __kfifo_len(display_q);
	
	trace_log_freep = TRACE_NEXT(trace_log_freep);
#endif
}

static void free_old_bufs(void)
{
	unsigned char i;

	/* We never free the last element in the queue, as it may 
	   still be needed by the display hardware */
	if (__kfifo_len(done_q) <= 1)
		return;

	__kfifo_get(done_q, &i, sizeof(i));

	ref_cnt[i]--;
	if (!ref_cnt[i])
		wake_up_interruptible(&blocked_q);
}

static void refire_timer(void)
{
    	if (is_framerate_changed)
	{	
	    	is_framerate_changed = 0;
		run_start = jiffies;
		run_len = 0;
	}
    	run_len++;
 	//flip_timer.expires = last_frame_timestamp + frame_delay;
 	flip_timer.expires = (unsigned long)(run_start + run_len * exact_frame_delay);

	if (time_before(flip_timer.expires, jiffies))
	{	
	    	/* we are already late for next frame, so set timer to paint it now */
	    	flip_timer.expires = run_start = jiffies;
		run_len = 0;
	}
	add_timer(&flip_timer);
}

static void flip_cb(unsigned long ignore)
{
	unsigned char next;
	
	if (!__kfifo_get(display_q, &next, sizeof(next)))
	    	return; /* should never occur */

	do_switch(next);
	last_frame_timestamp = jiffies;

	log_trace(next);

	free_old_bufs();
	__kfifo_put(done_q, &next, sizeof(next));

	if (!__kfifo_len(display_q))
		return;
	/* we have more buffers: refire timer */
	refire_timer();
}
	
static int n_avail_bufs(void)
{
	int i, n=0;

	for (i=DISP_START_BUF; i<disp_num_buffers; i++)
		n += !ref_cnt[i];

	return n;
}

/* Flowbee: check available disp buffers in scale mode. */
static int n_avail_scale_bufs(void)
{
	int i, n=0;

	for (i=0; i<DISP_YUVBUF_NUMS; i++)
		n += !ref_cnt[i];

	return n;
}

/* Flowbee: check available dec buffers. */
static int n_avail_dec_bufs(void)
{
	int i, n=0;

	for (i=0; i<dec_num_buffers; i++)
		n += !dec_ref_cnt[i];

	return n;
}

static void resume(unsigned long ignore)
{
	if (!is_paused)
		return;
	is_paused = 0;
	if (!__kfifo_len(display_q))
	    	return;
	run_len = 0;
	run_start = jiffies;
	flip_cb(0);
}

int arc_vsync_ioctl(struct inode *inode, struct file *file, u_int cmd,
				u_long arg, int con, struct fb_info *info)
{
	int i, j, res, len;
	unsigned char c;

	switch (cmd)
	{
	case ARCPGUFB_ALLOC:
#if 0	/*Flowbee@960125*/
		while (!n_avail_bufs()) {
			if (wait_event_interruptible(blocked_q, n_avail_bufs()))
				return -ERESTARTSYS;
		}
		for (i=0; i<CONFIG_ARCPGU_YUVBUFS && ref_cnt[i]; i++);
		ref_cnt[i]++;
		return put_user(i, (int *)arg);

#else
		i = 0x100;
		while (i-- > 0) {
			j = n_avail_bufs();
			if (!j){
				if (wait_event_interruptible(blocked_q, n_avail_bufs())){
					return -ERESTARTSYS;
				}
			}
			else{
				break;
			}
		}
		if ( j == 0){
			return -1;
		}
		for (i=DISP_START_BUF; i<disp_num_buffers && ref_cnt[i]; i++);
		ref_cnt[i]++;
		return put_user(i, (int *)arg);

#endif
		
	case ARCPGUFB_FREE:
		res = get_user(i, (int *)arg);
		if (res || i<DISP_START_BUF || i >= disp_num_buffers)
			return -EINVAL;
		ref_cnt[i]--;
		return 0;
		
	case ARCPGUFB_SEND:
		res = get_user(i, (int *)arg);
		if (res || i<DISP_START_BUF || i >= disp_num_buffers)
			return -EINVAL;

		ref_cnt[i]++;
		//printk("\nSend");		
		c = (unsigned char) i;
		__kfifo_put(display_q, &c, sizeof(c));
		if (timer_pending(&flip_timer) || is_paused)
		    	return 0;

		if (time_before_eq(last_frame_timestamp + (int)frame_delay, jiffies)) // XXX old non exact
		{
	    	/* already time for frame, so just display it */
	    	run_len = 0;
			run_start = jiffies;
	    	flip_cb(0);
			return 0;
		}
		refire_timer();
		return 0;
		
	/* Flowbee: Scale buffer. */
	case ARCPGUFB_SCALE_ALLOC:
	#if 1	/* Flowbee@960625. */
		i = 0x100;
		while (i-- > 0) {
			j = n_avail_scale_bufs();
			if (!j){
				if (wait_event_interruptible(blocked_q, n_avail_scale_bufs())){
					return -ERESTARTSYS;
				}
			}
			else{
				break;
			}
		}
		if ( j == 0){
			return -1;
		}
	#else
		while (!n_avail_scale_bufs()) {
			if (wait_event_interruptible(blocked_q, n_avail_scale_bufs()))
				return -ERESTARTSYS;
		}
	#endif
		
		for (i=0; i<DISP_YUVBUF_NUMS && ref_cnt[i]; i++);
		ref_cnt[i]++;
		//printk("\nAlloc");		
		return put_user(i, (int *)arg);
		
	case ARCPGUFB_SCALE_FREE:
		res = get_user(i, (int *)arg);
		if (res || i<0 || i >= DISP_YUVBUF_NUMS)
			return -EINVAL;
		ref_cnt[i]--;
		//printk("\nFree");
		return 0;
		
	case ARCPGUFB_SCALE_SEND:
		res = get_user(i, (int *)arg);
		if (res || i<0 || i >= DISP_YUVBUF_NUMS)
			return -EINVAL;

		ref_cnt[i]++;
		//printk("\nS_Send");
		c = (unsigned char) i;
		__kfifo_put(display_q, &c, sizeof(c));
		if (timer_pending(&flip_timer) || is_paused)
		    	return 0;
		if (time_before_eq(last_frame_timestamp + (int)frame_delay, jiffies)) // XXX old non exact
		{
	    	/* already time for frame, so just display it */
	    	run_len = 0;
			run_start = jiffies;
	    	flip_cb(0);
			return 0;
		}
		refire_timer();
		return 0;

	/* Flowbee: add decoder buffer. */		
	case ARCPGUFB_DEC_ALLOC:
	#if 1	/* Flowbee@960625. */
		i = 0x100;
		while (i-- > 0) {
			j = n_avail_dec_bufs();
			if (!j){
				if (wait_event_interruptible(blocked_q, n_avail_dec_bufs())){
					return -ERESTARTSYS;
				}
			}
			else{
				break;
			}
		}
		if ( j == 0){
			return -1;
		}
	#else
		while (!n_avail_dec_bufs()) {
			if (wait_event_interruptible(blocked_q, n_avail_dec_bufs()))
				return -ERESTARTSYS;
		}
	#endif

		for (i=0; i<dec_num_buffers&& dec_ref_cnt[i]; i++);
		dec_ref_cnt[i]++;
		//printk("\nDEC_ALLOC: %d(%d)",i, dec_ref_cnt[i]);
		return put_user(i, (int *)arg);
		
	case ARCPGUFB_DEC_FREE:
		res = get_user(i, (int *)arg);
		if (res || i<0 || i >= dec_num_buffers)
			return -EINVAL;
		dec_ref_cnt[i]--;
		//printk("\nDEC_FREE: %d(%d)",i, dec_ref_cnt[i]);		
		return 0;

	case ARCPGUFB_FRAMERATE_SET:
		res = get_user(i, (int *)arg);
		if (res)
		    	return res;
		if (i == 0)
		    	return -EINVAL;
		frame_rate = i/1000.0;
		frame_delay = HZ/frame_rate;
		exact_frame_delay = (double)HZ/frame_rate;
		if (!frame_delay)
		    frame_delay = 1;
		is_framerate_changed = 1;
		/* we could modify the timer here */
		return 0;
	case ARCPGUFB_QUEUE_LEN_GET:
		len = __kfifo_len(display_q);
		return put_user(len, (int *)arg);
	case ARCPGUFB_NUM_FREE_BUFS:
		len = n_avail_bufs();
		return put_user(len, (int *)arg);
	/* Flowbee: scale buffer. */		
	case ARCPGUFB_NUM_FREE_SCALE_BUFS:
		len = n_avail_scale_bufs();
		return put_user(len, (int *)arg);
	/* Flowbee: decode buffer. */
	case ARCPGUFB_NUM_FREE_DEC_BUFS:
		len = n_avail_dec_bufs();
		return put_user(len, (int *)arg);
		
	case ARCPGUFB_FLUSH:
		del_timer(&flip_timer);
		while (__kfifo_get(display_q, &c, sizeof(c)))
			ref_cnt[c]--;
		return 0;
	case ARCPGUFB_FRAME_DROP:
		del_timer(&flip_timer);
		if (__kfifo_get(display_q, &c, sizeof(c)))
			ref_cnt[c]--;
		if (__kfifo_len(display_q))
			add_timer(&flip_timer);
		return 0;
	case ARCPGUFB_PAUSE:
	case ARCPGUFB_PAUSE_MS:
		//printk("\nPaused");
		if (is_paused)
			return -EINVAL;
		is_paused = 1;
		del_timer(&flip_timer);
		if (cmd == ARCPGUFB_PAUSE)
			return 0;
		res = get_user(i, (int *)arg);
		if (res)
		    	return res;
	    	pause_timer.expires = jiffies + (i * HZ)/1000;
		add_timer(&pause_timer);
		return 0;
	case ARCPGUFB_RESUME:
		if (!is_paused)
			return -EINVAL;
		resume(0);
		return 0;
	case ARCPGUFB_RESET_SYNC:
		vsync_reset();
		return 0;
	default:
		return -EINVAL;
	}
}

void arc_vsync_init(frame_switch_cb_t fun)
{
#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *ent = create_proc_entry(PROC_NAME, 0, NULL);

	ent->read_proc = sync_read_cb;
#endif

	display_q = kfifo_alloc(CONFIG_ARCPGU_YUVBUFS, GFP_ATOMIC, NULL);
	done_q = kfifo_alloc(CONFIG_ARCPGU_YUVBUFS, GFP_ATOMIC, NULL);
	
	flip_timer.function = flip_cb;
	pause_timer.function = resume;
	do_switch = fun;
}

void arc_vsync_uninit(void)
{
	del_timer(&flip_timer);
	del_timer(&pause_timer);
#ifdef CONFIG_PROC_FS
	remove_proc_entry(PROC_NAME, NULL);
#endif
	kfifo_free(display_q);
	kfifo_free(done_q);
}

/* Flowbee test the image scalar. */
int scale_buf_alloc(void)
{
	int i, j;
	#if 1	/* Flowbee@960625. */
		i = 0x100;
		while (i-- > 0) {
			j = n_avail_scale_bufs();
			if (!j){
				if (wait_event_interruptible(blocked_q, n_avail_scale_bufs())){
					return -ERESTARTSYS;
				}
			}
			else{
				break;
			}
		}
		if ( j == 0){
			return -1;
		}
	#else
	while (!n_avail_scale_bufs()) {
		if (wait_event_interruptible(blocked_q, n_avail_scale_bufs()))
			return -ERESTARTSYS;
	}
	#endif
	
	for (i=0; i<DISP_YUVBUF_NUMS && ref_cnt[i]; i++);
	ref_cnt[i]++;
	//printk(" Alloc: %d", i);		
	return i;
}

int scale_buf_free(int i)
{
	if ( (i<0) || (i >= DISP_YUVBUF_NUMS))
		return -EINVAL;
	ref_cnt[i]--;
	//printk("\nFree: %d, ref: %d, %d, %d, %d\n", i, ref_cnt[0], ref_cnt[1], ref_cnt[2], ref_cnt[3]);
	return 0;
}

int scale_buf_send(int i)
{
	unsigned char c;
	if ( (i<0) || (i >= DISP_YUVBUF_NUMS) )
		return -EINVAL;

	ref_cnt[i]++;
	//printk("\n\nSend: %d", i);
	c = (unsigned char) i;
	__kfifo_put(display_q, &c, sizeof(c));
	if (timer_pending(&flip_timer) || is_paused)
    	return 0;
	if (time_before_eq(last_frame_timestamp + (int)frame_delay, jiffies)) // XXX old non exact
	{
    	/* already time for frame, so just display it */
    	run_len = 0;
		run_start = jiffies;
    	flip_cb(0);
		return 0;
	}
	refire_timer();
	return 0;
}


void dec_buf_alloc_free(unsigned char _dec_idx, unsigned char _free)
{
	if(_free == 1){
		dec_ref_cnt[_dec_idx]--;
		//printk("\ndec_free: %d(%d)",_dec_idx, dec_ref_cnt[_dec_idx]);
	}
	else{
		dec_ref_cnt[_dec_idx]++;
		//printk("\ndec_alloc: %d(%d)",_dec_idx, dec_ref_cnt[_dec_idx]);		
	}
}

void reset_mplayer_timer_queue(void)
{
	del_timer(&flip_timer);
	del_timer(&pause_timer);
	__kfifo_reset(display_q);
	__kfifo_reset(done_q);
	is_paused = 0;
	//-- Flowbee 2008-05-14 
	is_video_play = 0;	
	//-- Flowbee 2008-05-20 
	mplayer_status = MPLAYER_INIT;	
}

int check_video_paused(void)
{
	if(is_video_play == 1){
		return is_paused;
	}
	else{
		return -1;
	}
}
		
