#include "list.h"

/*------------------------------------------------------------------------------
Function name : ListInitList

Dexcription   : This routine initializes one list.
		
Input         : LIST_CONTAINER_ID pList
		
Return        : None
------------------------------------------------------------------------------*/
__inline void ListInitList(LIST_CONTAINER_ID pList)
{
	pList->Link.ForwardLink = pList->Link.BackwardLink = (LIST_ENTRY_ID) 0;
	pList->EntryCount = 0;
} // ListInitList


/*------------------------------------------------------------------------------
Function name : ListEmpty

Dexcription   : This routine checks if a list is empty.

Input         : LIST_CONTAINER_ID pList
		
Return        : TRUE/FALSE
------------------------------------------------------------------------------*/
__inline char ListEmpty(LIST_CONTAINER_ID pList)
{
	return(pList->Link.ForwardLink == (LIST_ENTRY_ID) 0);
} // ListEmpty


/*------------------------------------------------------------------------------
Function name : ListGetHead

Dexcription   : This routine returns the head of the list, but does not remove
		it.

Input         : LIST_CONTAINER_ID pList
		
Return        : LIST_ENTRY_ID ListHead
------------------------------------------------------------------------------*/
__inline LIST_ENTRY_ID ListGetHead(LIST_CONTAINER_ID pList)
{
	return pList->Link.ForwardLink;
} // ListGetHead


/*------------------------------------------------------------------------------
Function name : ListGetTail

Dexcription   : This routine returns the tail of the list, but does not remove
		it.

Input         : LIST_CONTAINER_ID pList
		
Return        : LIST_ENTRY_ID ListTail
------------------------------------------------------------------------------*/
__inline LIST_ENTRY_ID ListGetTail(LIST_CONTAINER_ID pList)
{
	return pList->Link.BackwardLink;
} // ListGetTail


/*------------------------------------------------------------------------------
Function name : ListPushHead

Dexcription   : This routine puts an element at the head of the list.

Input         : LIST_CONTAINER_ID pList
		LIST_ENTRY_ID     pEntry
		
Return        : None
------------------------------------------------------------------------------*/
__inline void ListPushHead(LIST_CONTAINER_ID pList, LIST_ENTRY_ID pEntry)
{
	pEntry->BackwardLink = (LIST_ENTRY_ID) 0;
	pEntry->ForwardLink = pList->Link.ForwardLink;

	if(pList->Link.ForwardLink == (LIST_ENTRY_ID) 0) {
		pList->Link.BackwardLink = (LIST_ENTRY_ID) pEntry;
	}
	else{
		pList->Link.ForwardLink->BackwardLink = pEntry;
	}

	pList->Link.ForwardLink = (LIST_ENTRY_ID) pEntry;

	pList->EntryCount++;
} // ListPushHead


/*------------------------------------------------------------------------------
Function name : ListPushTail

Dexcription   : This routine puts an element at the tail of the list.

Input         : LIST_CONTAINER_ID pList
		LIST_ENTRY_ID     pEntry
		
Return        : None
------------------------------------------------------------------------------*/
__inline void ListPushTail(LIST_CONTAINER_ID pList, LIST_ENTRY_ID pEntry)
{
	pEntry->BackwardLink = pList->Link.BackwardLink;

	if(pList->Link.BackwardLink) {
		pList->Link.BackwardLink->ForwardLink = pEntry;
	}
	else {
		pList->Link.ForwardLink = pEntry;
	}

	pList->Link.BackwardLink = pEntry;
	pEntry->ForwardLink = (LIST_ENTRY_ID) 0;

	pList->EntryCount++;
} // ListPushTail


/*------------------------------------------------------------------------------
Function name : ListPopHead

Dexcription   : This routine pops the head of the list and this element will
		be removed from this list at the same time.

Input         : LIST_CONTAINER_ID pList
		
Return        : LIST_ENTRY_ID pEntry;
------------------------------------------------------------------------------*/
__inline LIST_ENTRY_ID ListPopHead(LIST_CONTAINER_ID pList)
{
	LIST_ENTRY_ID pEntry;

	pEntry = pList->Link.ForwardLink;

	if(pEntry) {
		pList->Link.ForwardLink = pEntry->ForwardLink;
		if(pList->Link.ForwardLink == (LIST_ENTRY_ID) 0) {
			pList->Link.BackwardLink = (LIST_ENTRY_ID) 0;
		}
		else {
			pList->Link.ForwardLink->BackwardLink = (LIST_ENTRY_ID) 0;
		}

		pList->EntryCount--;
	}
    
	return pEntry;
} // ListPopHead


/*------------------------------------------------------------------------------
Function name : ListPopTail

Dexcription   : This routine pops the tail of the list and this element will
		be removed from this list at the same time.

Input         : LIST_CONTAINER_ID pList
		
Return        : LIST_ENTRY_ID pEntry;
------------------------------------------------------------------------------*/
__inline LIST_ENTRY_ID ListPopTail(LIST_CONTAINER_ID pList)
{
	LIST_ENTRY_ID pEntry;

	pEntry = pList->Link.BackwardLink;

	if(pEntry) {
		pList->Link.BackwardLink = pEntry->BackwardLink;
		if(pList->Link.BackwardLink == (LIST_ENTRY_ID) 0) {
			pList->Link.ForwardLink = (LIST_ENTRY_ID) 0;
		}
		else {
			pList->Link.BackwardLink->ForwardLink = (LIST_ENTRY_ID) 0;
		}

		pList->EntryCount--;
	}
    
	return pEntry;
} // ListPopTail


/*------------------------------------------------------------------------------
Function name : ListNextEntry

Dexcription   : This routine gets next element beyond the input element.

Input         : LIST_CONTAINER_ID pList
		
Return        : LIST_ENTRY_ID NextEntry
------------------------------------------------------------------------------*/
__inline LIST_ENTRY_ID ListNextEntry(LIST_ENTRY_ID pEntry)
{
	return pEntry->ForwardLink;
} // ListNextEntry


/*------------------------------------------------------------------------------
Function name : ListPrevEntry

Dexcription   : This routine gets previous element before the input element.

Input         : LIST_CONTAINER_ID pList
		
Return        : LIST_ENTRY_ID PreviousEntry
------------------------------------------------------------------------------*/
__inline LIST_ENTRY_ID ListPrevEntry(LIST_ENTRY_ID pEntry)
{
	return pEntry->BackwardLink;
} // ListPrevEntry


/*------------------------------------------------------------------------------
Function name : ListOk

Dexcription   : This routine check the links and counters are consistent.

Input         : LIST_CONTAINER_ID pList
		
Return        : TRUE/FALSE
------------------------------------------------------------------------------*/
__inline char ListOk(LIST_CONTAINER_ID pList)
{
	LIST_ENTRY_ID pEntry;
	unsigned long Count;

	Count = 0;
	pEntry = ListGetHead(pList);
	while(pEntry) {
		Count++;
		pEntry = ListNextEntry(pEntry);
	}

	if(Count != pList->EntryCount) {
		return 0;
	}

	Count = 0;
	pEntry = ListGetTail(pList);
	while(pEntry) {
		Count++;
		pEntry = ListPrevEntry(pEntry);
	}

	if(Count != pList->EntryCount) {
		return 0;
	}

	return 1;
} // ListOk


/*------------------------------------------------------------------------------
Function name : ListGetSize

Dexcription   : This routine gets the number of elements in the list.

Input         : LIST_CONTAINER_ID pList
		
Return        : unsigned long EntryCount
------------------------------------------------------------------------------*/
__inline unsigned long ListGetSize(LIST_CONTAINER_ID pList)
{
	return pList->EntryCount;
} // ListGetSize

