#include "tiff.h"
//------------------------------------------------------------------------------------------------
static void tiff_get_ascii(FILE *fn, unsigned char *buf, int offset, int size)
{
    fseek(fn, offset, SEEK_SET);
    memset(buf, 0x00, size);
    fread(buf, size, 1, fn);
}

//------------------------------------------------------------------------------------------------
static unsigned short tiff_get_short(unsigned char **buf, char endian)
{
    unsigned short ret=0;
    unsigned char *cur_ptr = *buf;

    switch(endian)
    {
        case TIFF_BIG_ENDIAN:
            ret = ((cur_ptr[0] << 8) | cur_ptr[1]);
            break;
        case TIFF_LITTLE_ENDIAN:
            ret = ((cur_ptr[1] << 8) | cur_ptr[0]);
            break;

        default:
        break;
    }
    *buf+=2;
    return ret;
}

//------------------------------------------------------------------------------------------------
static unsigned int tiff_get_int(unsigned char **buf, char endian)
{
    unsigned int ret=0;
    unsigned char *cur_ptr = *buf;

    switch(endian)
    {
        case TIFF_BIG_ENDIAN:
            ret = ((cur_ptr[0] << 24) | (cur_ptr[1] << 16) | (cur_ptr[2] << 8) | cur_ptr[3]);
            break;
        case TIFF_LITTLE_ENDIAN:
            ret = ((cur_ptr[3] << 24) | (cur_ptr[2] << 16) | (cur_ptr[1] << 8) | cur_ptr[0]);
            break;

        default:
        break;
    }
    *buf+=4;
    return ret;
}

//------------------------------------------------------------------------------------------------
static unsigned int tiff_get_index(unsigned char **buf, char endian, int type)
{
    unsigned int ret=0;
    unsigned char *cur_ptr = *buf;

    switch(type)
    {
        case TIFF_SHORT:
        case TIFF_SSHORT:
            ret = tiff_get_short(buf, endian);
            break;
        default:
            ret = tiff_get_int(buf, endian);
            break;
    }

    return ret;
}

//------------------------------------------------------------------------------------------------
/*
 * Tiff Header Structure
 * Reference : 1. http://partners.adobe.com/public/developer/en/tiff/TIFF6.pdf
 *             2. libtiff-3.8.2
 * 
 *             Header                            Directory Entry
 *          |---------|                          |-----------| 
 *        0 |         | Byte Order           X   |           | Tag
 *          |---------|                          |-----------|
 *        2 |         | 42                   X+2 |           | Type
 *          |---------|                          |-----------|
 *        4 |         | Offset of 0th IFD    X+4 |           | Count
 *          | -  A  - |                          | - - - - - | 
 *        6 |         |                          |           |
 *          |---------|                      X+8 |-----------|
 *               |                               |           | Value or Offset
 *               |                               | - - - - - |
 *               V                               |           |
 *                                               |-----------|
 *              IFD                                    |
 *          |---------|                                |
 * A        |    B    | Number of Directory Entries    V
 *          |---------|                          |-----------|
 * A+2      |         | Directory Entry 0        |           | Value
 *          |---------|                          |-----------|
 * A+14     |         | Directory Entry 1
 *          |---------|
 * A+26     |         | Directory Entry 2
 *          |---------|
 *          |         |
 *          |---------|
 * A+2+B*12 |         | Offset of next IFD
 *          | - - - - |
 *          |         |
 *          |---------|
 *
 */
struct tiff_hdr * tiff_header(char *filestr)
{
    struct tiff_hdr *new_tiff_hdr;
    FILE *fn;
    unsigned char *ascii_buf;
    unsigned int i;
    unsigned char *filebuf, *cur_buf_ptr;
    char endian_type;
    unsigned short *tiff_magic_ptr, tiff_version;
    unsigned int IFD_offset=0;
    unsigned int NumberOfDirectoryEntries=0;
    unsigned short DirEntryTag=0, DirEntryType=0;
    unsigned int DirEntryCount=0, DirEntryValueOffset=0;
    unsigned int nsamples=0, planarconfig=0;
    unsigned int is_ycbcr_format=0;
   
    filebuf = malloc(MAX_READ_SIZE);
    if (filebuf == 0)
    {
        return 0;
    }

    fn = fopen(filestr, "r");
    if (fn == 0)
    {
        free(filebuf);
        return 0;
    }

    //read 4096 bytes
    fread(filebuf, MAX_READ_SIZE, 1, fn);
   
    //check tiff magic number
    tiff_magic_ptr = (unsigned short *)&filebuf[0];
    if (*tiff_magic_ptr == 0x4949)
    {
        TIF_DBG("tiff magic = II\n");
        endian_type = TIFF_LITTLE_ENDIAN;
    }
    else if (*tiff_magic_ptr == 0x4d4d)
    {
        TIF_DBG("tiff magic = MM\n");
        endian_type = TIFF_BIG_ENDIAN;
    }
    else
    {
        TIF_DBG("tiff magic = 0x%x, not tiff file\n", *tiff_magic_ptr);
        free(filebuf);
        fclose(fn);
        return 0;
    }

    cur_buf_ptr = &filebuf[2];
    tiff_version = tiff_get_short(&cur_buf_ptr, endian_type);
    TIF_DBG("tiff_version = 0x%x(%d)\n", tiff_version, tiff_version);

    IFD_offset = tiff_get_int(&cur_buf_ptr, endian_type);
    TIF_DBG("IFD_offset = 0x%x(%d)\n", IFD_offset, IFD_offset);

    //seek fn to IFD_offset ,and update filebuf, 4096bytes
    fseek(fn, IFD_offset, SEEK_SET);
    fread(filebuf, MAX_READ_SIZE, 1, fn);
    cur_buf_ptr = &filebuf[0];

    //get Number of Directory Entries
    NumberOfDirectoryEntries = tiff_get_short(&cur_buf_ptr, endian_type);
    TIF_DBG("NumberOfDirectoryEntries = 0x%x(%d)\n", NumberOfDirectoryEntries, NumberOfDirectoryEntries);

    new_tiff_hdr = malloc(sizeof(struct tiff_hdr));
    if (new_tiff_hdr == 0)
    {
        free(filebuf);
        fclose(fn);
        return 0;
    }

    memset(new_tiff_hdr, 0x00, sizeof(struct tiff_hdr));
    new_tiff_hdr->compression_type = COMPRESSION_NONE;
//Robert 20080410 Set Default planarconfig
    planarconfig = PLANARCONFIG_CONTIG;   //set default planarconfig
   
    //Process Directory
    for (i=0; i<NumberOfDirectoryEntries && i < 300; i++)
    {
#if DEBUG_TIF_PARSER
        int dump_info = 1;
#endif
        TIF_DBG("===== Directory Entry (%d) ======\n", i);
        DirEntryTag = tiff_get_short(&cur_buf_ptr, endian_type);
        DirEntryType = tiff_get_short(&cur_buf_ptr, endian_type);
        DirEntryCount = tiff_get_int(&cur_buf_ptr, endian_type);
//Robert 20080410 Fix DirEntryValueOffset check for BigEndian
#if 0
        DirEntryValueOffset = tiff_get_int(&cur_buf_ptr, endian_type);
#else
        if (((DirEntryType == TIFF_SHORT) || (DirEntryType == TIFF_SSHORT)) && (DirEntryCount == 1))
        {
            DirEntryValueOffset = tiff_get_short(&cur_buf_ptr, endian_type);
            cur_buf_ptr += 2;	//need 4 bytes consume
        }
        else
        {
            DirEntryValueOffset = tiff_get_int(&cur_buf_ptr, endian_type);
        }
#endif

        switch(DirEntryTag)
        {
            case TIFFTAG_COMPRESSION:
                new_tiff_hdr->compression_type = DirEntryValueOffset;
                TIF_DBG("It's TIFFTAG_COMPRESSION!!!!\n");
                TIF_DBG("compression_type = %d\n", new_tiff_hdr->compression_type);
#if DEBUG_TIF_PARSER
                switch (new_tiff_hdr->compression_type)
                {
                    case COMPRESSION_NONE:
                        TIF_DBG("NO Compression, RAW data\n");
                        break;
                    case COMPRESSION_JPEG:
                        TIF_DBG("JPEG Compression\n");
                        break;
                    case COMPRESSION_PACKBITS:
                        TIF_DBG("Macintosh Packbits Compression\n");
                        break;
                    case COMPRESSION_LZW:
                        TIF_DBG("LZW Compression, but we don't support it\n");
                        break;
                    case COMPRESSION_DEFLATE:
                        TIF_DBG("Deflate Compression, but we don't support it\n");
                        break;
                    default:
                        TIF_DBG("Unknown Compression Format, no support\n");
                        break;
                }
#endif
                break;
            case TIFFTAG_IMAGEWIDTH:
                new_tiff_hdr->width = DirEntryValueOffset;
                TIF_DBG("imagewidth = %d\n", new_tiff_hdr->width);
                break;
            case TIFFTAG_IMAGELENGTH:
                new_tiff_hdr->height = DirEntryValueOffset;
                TIF_DBG("imageheight = %d\n", new_tiff_hdr->height);
                break;
            case TIFFTAG_DATETIME:
                TIF_DBG("It's TIFFTAG_DATETIME!!\n");

                ascii_buf = malloc(DirEntryCount);
                tiff_get_ascii(fn, ascii_buf, DirEntryValueOffset, DirEntryCount);
                TIF_DBG("strlen(ascii_buf)=%d, ascii_buf=%s!\n", strlen(ascii_buf), ascii_buf);
                snprintf(new_tiff_hdr->DateTime, 20, "%s", ascii_buf); 
                free(ascii_buf);
                break;
            case TIFFTAG_MAKE:
                TIF_DBG("It's TIFFTAG_MAKE!!\n");

                ascii_buf = malloc(DirEntryCount);
                tiff_get_ascii(fn, ascii_buf, DirEntryValueOffset, DirEntryCount);
                TIF_DBG("strlen(ascii_buf)=%d, ascii_buf=%s!\n", strlen(ascii_buf), ascii_buf);
                snprintf(new_tiff_hdr->CameraMake, 32, "%s", ascii_buf); 
                free(ascii_buf);
                break;
            case TIFFTAG_MODEL:
                TIF_DBG("It's TIFFTAG_MODEL!!\n");

                ascii_buf = malloc(DirEntryCount);
                tiff_get_ascii(fn, ascii_buf, DirEntryValueOffset, DirEntryCount);
                TIF_DBG("strlen(ascii_buf)=%d, ascii_buf=%s!\n", strlen(ascii_buf), ascii_buf);
                snprintf(new_tiff_hdr->CameraModel, 40, "%s", ascii_buf); 

                free(ascii_buf);
                break;
            case TIFFTAG_SAMPLESPERPIXEL:
                nsamples = DirEntryValueOffset;
                break;
            case TIFFTAG_PLANARCONFIG:
                planarconfig = DirEntryValueOffset;
                break;
            case TIFFTAG_YCBCRSUBSAMPLING:	//N = 2, val = Cb & Cr
            case TIFFTAG_YCBCRPOSITIONING:
                if (DirEntryValueOffset > 0)
                {
                    is_ycbcr_format = 1;
                }
                break;

            default:
#if DEBUG_TIF_PARSER
                dump_info = 0;
#endif
                break;
        }

#if DEBUG_TIF_PARSER
        if (dump_info)
        {
            TIF_DBG("Tag:         0x%x(%d)\n", DirEntryTag, DirEntryTag);
            TIF_DBG("Type:        0x%x(%d)\n", DirEntryType, DirEntryType);
            TIF_DBG("Count:       0x%x(%d)\n", DirEntryCount, DirEntryCount);
            TIF_DBG("ValueOffset: 0x%x(%d)\n", DirEntryValueOffset, DirEntryValueOffset);
        }
#endif

    }

    fclose(fn);
    free(filebuf);

//Robert 20080410 Add is_ycbcr_format == 0 check

//    if ( !( (new_tiff_hdr->compression_type == COMPRESSION_NONE || 
//          new_tiff_hdr->compression_type == COMPRESSION_JPEG ||
//          new_tiff_hdr->compression_type == COMPRESSION_PACKBITS)
    if ( !( (new_tiff_hdr->compression_type == COMPRESSION_NONE )
          && (new_tiff_hdr->width > 0) 
          && (new_tiff_hdr->height > 0)
          && (nsamples == 3)
          && (planarconfig == PLANARCONFIG_CONTIG)
          && (is_ycbcr_format == 0) ) )
    {
        free(new_tiff_hdr);
        new_tiff_hdr = 0;
    }

    return new_tiff_hdr;
}


//------------------------------------------------------------------------------------------------
void *GetTiffHdr(char* filename)
{
	return tiff_header(filename);
}
//------------------------------------------------------------------------------------------------
void FreeTiffHdr(void* data)
{
	free(data);
}



#if 0
void main(int argc, char **argv)
{
    int i;
    TIF_DBG("argc=%d, argv[1]=%s!\n", argc, argv[1]);
    for (i=1; i<argc; i++)
    {
        TIF_DBG("==============================\n");
        TIF_DBG("file : %s\n", argv[i]);
        tiff_header(argv[i]);
    }
}
#endif
